import atexit
import multiprocessing

import click
import gevent
import strictyaml
import zmq.green as zmq
from loguru import logger
from playhouse.apsw_ext import APSWDatabase

from reprobench.core.db import Run, db
from reprobench.core.events import RUN_COMPLETE
from reprobench.core.observers import CoreObserver
from reprobench.core.schema import schema
from reprobench.utils import clean_up, get_db_path, import_class, read_config

BACKEND_ADDRESS = "inproc://backend"


class BenchmarkServer:
    def __init__(self, observers, db_path, **kwargs):
        super().__init__()
        db.initialize(APSWDatabase(db_path))
        self.frontend_address = kwargs.pop("address", "tcp://*:31334")
        self.serve_forever = kwargs.pop("forever", False)
        self.observers = observers + [CoreObserver]
        self.jobs_waited = Run.select().where(Run.status < Run.DONE).count()

    def loop(self):
        while True:
            address, event_type, payload = self.frontend.recv_multipart()
            logger.debug((address, event_type, payload))
            self.backend.send_multipart([event_type, payload, address])
            if event_type == RUN_COMPLETE:
                self.jobs_waited -= 1
            if not self.serve_forever and self.jobs_waited == 0:
                break

    def run(self):
        self.context = zmq.Context()
        self.frontend = self.context.socket(zmq.ROUTER)
        self.frontend.bind(self.frontend_address)
        self.backend = self.context.socket(zmq.PUB)
        self.backend.bind(BACKEND_ADDRESS)

        logger.info(f"Listening on {self.frontend_address}")

        observer_greenlets = []
        for observer in self.observers:
            greenlet = gevent.spawn(
                observer.observe,
                self.context,
                backend_address=BACKEND_ADDRESS,
                frontend=self.frontend,
            )
            observer_greenlets.append(greenlet)

        serverlet = gevent.spawn(self.loop)
        serverlet.join()

        gevent.killall(observer_greenlets)


@click.command(name="server")
@click.option("-c", "--config", required=True, type=click.Path())
@click.option("-f", "--forever", help="Serve forever", is_flag=True)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(file_okay=False, writable=True, resolve_path=True),
    default="./output",
    required=True,
    show_default=True,
)
@click.argument("server_address")
def cli(config, output_dir, server_address):
    atexit.register(clean_up)

    config = read_config(config)
    database = get_db_path(output_dir)

    observers = []
    for observer in config["observers"]:
        observers.append(import_class(observer["module"]))

    server = BenchmarkServer(observers, database, address=server_address)
    server.run()


if __name__ == "__main__":
    cli()



from loguru import logger

from reprobench.core.base import Step, Observer
from reprobench.executors.events import STORE_RUNSTATS
from reprobench.utils import recv_event

from .db import RunStatistic


class RunStatisticObserver(Observer):
    SUBSCRIBED_EVENTS = [STORE_RUNSTATS]

    @classmethod
    def handle_event(cls, event_type, payload, **kwargs):
        if event_type == STORE_RUNSTATS:
            RunStatistic.create(**payload)


class Executor(Step):
    @classmethod
    def register(cls, config={}):
        RunStatistic.create_table()




import atexit
import os
from signal import signal

import zmq
from loguru import logger
from playhouse.apsw_ext import APSWDatabase

from reprobench.core.db import Run, db
from reprobench.core.events import RUN_COMPLETE, RUN_REGISTER
from reprobench.utils import clean_up, decode_message, import_class, send_event


def execute(args):
    atexit.register(clean_up)

    run_id, config, server_address = args

    context = zmq.Context()
    socket = context.socket(zmq.DEALER)
    socket.connect(server_address)
    send_event(socket, RUN_REGISTER, run_id)
    run = decode_message(socket.recv())
    tool = import_class(run["tool"])
    context = config.copy()
    context["tool"] = tool
    context["run"] = run
    context["socket"] = socket
    logger.info(f"Processing task: {run['directory']}")

    for runstep in config["steps"]["run"]:
        logger.debug(f"Running step {runstep['module']}")
        step = import_class(runstep["module"])
        step.execute(context, runstep.get("config", {}))

    send_event(socket, RUN_COMPLETE, run_id)



import atexit
import itertools
import os
import shutil
import signal
import time
from datetime import datetime
from multiprocessing import Process
from multiprocessing.pool import Pool
from pathlib import Path

import click
from loguru import logger
from playhouse.apsw_ext import APSWDatabase
from tqdm import tqdm

from reprobench.core.base import Runner
from reprobench.core.bootstrap import bootstrap
from reprobench.core.db import Run, db
from reprobench.core.server import BenchmarkServer
from reprobench.task_sources.local import LocalSource
from reprobench.utils import get_db_path, import_class, init_db, read_config

from .worker import execute


class LocalRunner(Runner):
    def __init__(self, config, **kwargs):
        self.config = config
        self.output_dir = kwargs.pop("output_dir", "./output")
        self.resume = kwargs.pop("resume", False)
        self.server_address = kwargs.pop("server", "tcp://127.0.0.1:31313")
        self.observers = []
        self.queue = []

    def exit(self):
        if len(self.queue) > 0 and hasattr(self, "pool"):
            self.pool.terminate()
            self.pool.join()

        if not self.resume and not self.setup_finished:
            shutil.rmtree(self.output_dir)

    def populate_unfinished_runs(self):
        query = Run.select(Run.id).where(Run.status < Run.DONE)
        self.queue = [(run.id, self.config, self.server_address) for run in query]

    def run(self):
        init_db(get_db_path(self.output_dir))
        self.populate_unfinished_runs()
        db.close()

        if len(self.queue) == 0:
            logger.success("No tasks remaining to run")
            exit(0)

        logger.debug("Executing runs...")

        with Pool() as pool:
            it = pool.imap_unordered(execute, self.queue)
            progress_bar = tqdm(desc="Executing runs", total=len(self.queue))
            for _ in it:
                progress_bar.update()
            progress_bar.close()

        logger.debug("Running teardown on all tools...")
        for tool in self.config["tools"].values():
            import_class(tool["module"]).teardown()


@click.command("local")
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(file_okay=False, writable=True, resolve_path=True),
    default="./output",
    show_default=True,
)
@click.option("-r", "--resume", is_flag=True)
@click.option("-s", "--server", default="tcp://127.0.0.1:31313")
@click.argument("config", type=click.Path())
def cli(config, output_dir, **kwargs):
    config = read_config(config)
    runner = LocalRunner(config, output_dir=output_dir, **kwargs)
    runner.run()


if __name__ == "__main__":
    cli()



import atexit
import os
import signal
import time

import click
import zmq
from loguru import logger

from reprobench.core.schema import schema
from reprobench.runners.local.worker import RUN_REGISTER
from reprobench.utils import (
    clean_up,
    decode_message,
    import_class,
    send_event,
    read_config,
)


@click.command()
@click.option("-c", "--config", required=True, type=click.Path())
@click.argument("server_address")
@click.argument("run_id", type=int)
def run(config, server_address, run_id):
    atexit.register(clean_up)

    config = read_config(config)

    context = zmq.Context()
    socket = context.socket(zmq.DEALER)
    socket.connect(server_address)
    send_event(socket, RUN_REGISTER, run_id)
    run = decode_message(socket.recv())

    tool = import_class(run["tool"])
    context = config.copy()
    context["socket"] = socket
    context["tool"] = tool
    context["run"] = run
    logger.info(f"Processing task: {run['directory']}")

    for runstep in config["steps"]["run"]:
        logger.debug(f"Running step {runstep['module']}")
        step = import_class(runstep["module"])
        step.execute(context, runstep.get("config", {}))


if __name__ == "__main__":
    run()



import subprocess
from pathlib import Path

from loguru import logger
from psmon import ProcessMonitor
from psmon.limiters import CpuTimeLimiter, MaxMemoryLimiter, WallTimeLimiter

from reprobench.core.events import RUN_FINISH, RUN_START
from reprobench.utils import send_event

from .base import Executor
from .db import RunStatistic
from .events import STORE_RUNSTATS


class PsmonExecutor(Executor):
    @classmethod
    def execute(cls, context, config={}):
        tool = context["tool"]
        limits = context["limits"]
        run_id = context["run"]["id"]
        tool.pre_run(context)

        cwd = context["run"]["directory"]
        out_file = (Path(cwd) / "run.out").open("wb")
        err_file = (Path(cwd) / "run.err").open("wb")

        cmd = tool.cmdline(context)
        logger.debug(f"Running {cwd}")
        logger.trace(cmd)

        monitor = ProcessMonitor(
            cmd, cwd=cwd, stdout=out_file, stderr=err_file, freq=15
        )
        monitor.subscribe("wall_time", WallTimeLimiter(limits["time"] + 15))
        monitor.subscribe("cpu_time", CpuTimeLimiter(limits["time"]))
        monitor.subscribe("max_memory", MaxMemoryLimiter(limits["memory"]))

        send_event(context["socket"], RUN_START, run_id)
        stats = monitor.run()
        send_event(context["socket"], RUN_FINISH, run_id)

        logger.debug(f"Finished {cwd}")

        verdict = None
        if stats["error"] == TimeoutError:
            verdict = RunStatistic.TIMEOUT
        elif stats["error"] == MemoryError:
            verdict = RunStatistic.MEMOUT
        elif stats["error"] or stats["return_code"] != 0:
            verdict = RunStatistic.RUNTIME_ERR
        else:
            verdict = RunStatistic.SUCCESS

        del stats["error"]

        payload = dict(run_id=run_id, verdict=verdict, **stats)
        send_event(context["socket"], STORE_RUNSTATS, payload)

        tool.post_run(context)



import itertools
import os
import signal
import subprocess
import time
from multiprocessing.pool import Pool
from pathlib import Path
from string import Template

import click
from loguru import logger
from playhouse.apsw_ext import APSWDatabase

from reprobench.core.base import Runner
from reprobench.core.bootstrap import bootstrap
from reprobench.core.db import Run, db
from reprobench.utils import get_db_path, import_class, init_db, read_config

from .utils import create_ranges

DIR = os.path.dirname(os.path.realpath(__file__))


class SlurmRunner(Runner):
    def __init__(self, config_path, python_path, server_address, **kwargs):
        self.config = read_config(config_path)
        self.config_path = config_path
        self.python_path = python_path
        self.server_address = server_address
        self.output_dir = kwargs.pop("output_dir", "./output")
        self.resume = kwargs.pop("resume", False)
        self.templates = {}
        self.templates["server"] = kwargs.pop(
            "server_template_file", os.path.join(DIR, "./slurm.server.job.tpl")
        )
        self.templates["run"] = kwargs.pop(
            "run_template_file", os.path.join(DIR, "./slurm.run.job.tpl")
        )
        self.templates["compile"] = kwargs.pop(
            "compile_template_file", os.path.join(DIR, "./slurm.compile.job.tpl")
        )
        self.queue = []

    def populate_unfinished_runs(self):
        query = Run.select(Run.id).where(Run.status < Run.DONE)
        self.queue = [run.id for run in query]

    def generate_template(self, template_type):
        template_file = self.templates[template_type]
        with open(template_file) as tpl:
            template = Template(tpl.read())
            job_str = template.safe_substitute(
                output_dir=self.output_dir,
                mem=int(1 + self.config["limits"]["memory"] / 1024 / 1024),  # mb
                time=int(1 + (self.config["limits"]["time"] + 15) / 60),  # minutes
                run_ids=create_ranges(self.queue),
                python_path=self.python_path,
                config_path=self.config_path,
                server_address=self.server_address,
            )

        job_path = Path(self.output_dir) / f"slurm.{template_type}.job"
        with open(job_path, "w") as job:
            job.write(job_str)

        return str(job_path.resolve())

    def run(self):
        init_db(get_db_path(self.output_dir))
        self.populate_unfinished_runs()
        db.close()

        if len(self.queue) == 0:
            logger.success("No tasks remaining to run")
            exit(0)

        logger.debug("Generating templates...")
        templates = {t: self.generate_template(t) for t in ["run"]}

        logger.info("Submitting jobs to SLURM...")

        run_cmd = ["sbatch", "--parsable", templates["run"]]
        run_job = subprocess.check_output(run_cmd).decode().strip()
        logger.debug(f"Run job id: {run_job}")


@click.command("slurm")
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(file_okay=False, writable=True, resolve_path=True),
    default="./output",
    required=True,
    show_default=True,
)
@click.option("--run-template", type=click.Path(dir_okay=False, resolve_path=True))
@click.option("--compile-template", type=click.Path(dir_okay=False, resolve_path=True))
@click.option("-r", "--resume", is_flag=True)
@click.option("-p", "--python-path", required=True, type=click.Path(resolve_path=True))
@click.option("-s", "--server", required=True)
@click.argument("config", type=click.Path())
def cli(config, output_dir, python_path, server, **kwargs):
    runner = SlurmRunner(
        config_path=config, python_path=python_path, server_address=server, **kwargs
    )
    runner.run()


if __name__ == "__main__":
    cli()



import zmq.green as zmq

from reprobench.utils import recv_event


class Runner:
    def __init__(self, config):
        pass

    def run(self):
        pass


class Observer:
    SUBSCRIBED_EVENTS = []

    @classmethod
    def observe(cls, context, backend_address, frontend):
        socket = context.socket(zmq.SUB)
        socket.connect(backend_address)

        for event in cls.SUBSCRIBED_EVENTS:
            socket.setsockopt(zmq.SUBSCRIBE, event)

        while True:
            event_type, payload, address = recv_event(socket)
            cls.handle_event(event_type, payload, frontend=frontend, address=address)

    @classmethod
    def handle_event(cls, event_type, payload, **kwargs):
        pass


class Step:
    @classmethod
    def register(cls, config={}):
        pass

    @classmethod
    def execute(cls, context, config={}):
        pass


class Tool:
    name = "Base Tool"
    REQUIRED_PATHS = []

    @classmethod
    def setup(cls):
        pass

    @classmethod
    def version(cls):
        return "1.0.0"

    @classmethod
    def pre_run(cls, context):
        pass

    @classmethod
    def cmdline(cls, context):
        pass

    @classmethod
    def post_run(cls, context):
        pass

    @classmethod
    def teardown(cls):
        pass



import msgpack
import zmq.green as zmq
from loguru import logger

from reprobench.core.db import Run
from reprobench.core.base import Observer
from reprobench.core.events import RUN_FINISH, RUN_REGISTER, RUN_START
from reprobench.utils import decode_message, encode_message, recv_event


class CoreObserver(Observer):
    SUBSCRIBED_EVENTS = [RUN_REGISTER, RUN_START, RUN_FINISH]

    @classmethod
    def handle_event(cls, event_type, payload, **kwargs):
        frontend = kwargs.pop("frontend")
        address = kwargs.pop("address")

        if event_type == RUN_REGISTER:
            run = Run.get(payload)
            run_dict = dict(
                id=run.id,
                task=run.task_id,
                tool=run.tool_id,
                directory=run.directory,
                parameters=list(run.parameter_group.parameters.dicts()),
            )
            frontend.send_multipart([address, encode_message(run_dict)])
        elif event_type == RUN_START:
            Run.update(status=Run.RUNNING).where(Run.id == payload).execute()
        elif event_type == RUN_FINISH:
            Run.update(status=Run.DONE).where(Run.id == payload).execute()



#!/usr/bin/env python

import argparse
import os
import sys

import click
import strictyaml
from loguru import logger

from reprobench.core.bootstrap import cli as bootstrap_cli
from reprobench.core.server import cli as server_cli
from reprobench.core.schema import schema
from reprobench.utils import import_class, read_config

from reprobench.runners import cli as runner_cli


@click.group()
@click.version_option()
@click.option("--verbose", "-v", "verbosity", count=True, default=0, help="Verbosity")
def cli(verbosity):
    sys.path.append(os.getcwd())
    logger.remove()

    if verbosity == 0:
        logger.add(sys.stderr, level="ERROR")
    elif verbosity == 1:
        logger.add(sys.stderr, level="WARNING")
    elif verbosity == 2:
        logger.add(sys.stderr, level="INFO")
    elif verbosity == 3:
        logger.add(sys.stderr, level="DEBUG")
    elif verbosity >= 4:
        logger.add(sys.stderr, level="TRACE")


cli.add_command(bootstrap_cli)
cli.add_command(server_cli)
cli.add_command(runner_cli)


if __name__ == "__main__":
    cli()
